import os

import pandas as pd


try:
    module_path = os.path.abspath(os.path.join("."))
    us_cities = pd.read_csv(os.path.join(module_path, "data/us_cities.csv"))
except FileNotFoundError:
    module_path = os.path.abspath(os.path.join(".."))
    us_cities = pd.read_csv(os.path.join(module_path, "data/us_cities.csv"))

us_cities.rename(columns={"LAT": "lat", "LON": "lon"}, inplace=True)

us_cities = us_cities[(us_cities["State"] != "AK") & (us_cities["State"] != "HI")]

us_cities["lon"] = -us_cities["lon"]

us_cities.reset_index(drop=True, inplace=True)
us_cities.index += 1

us_cities_dict = us_cities.to_dict(orient="index")



import plotly.graph_objects as go
import numpy as np


def add_trace(
    fig, cities_dict, city_1, city_2, lon=lambda x: x["lon"], lat=lambda x: x["lat"]
):

    city_1_lon = lon(cities_dict[city_1])
    city_1_lat = lat(cities_dict[city_1])
    city_2_lon = lon(cities_dict[city_2])
    city_2_lat = lat(cities_dict[city_2])

    if (
        city_1_lon < 0 < city_2_lon
        and np.abs(city_1_lon) > 90
        and np.abs(city_2_lon) > 90
    ):
        city_1_lon = city_1_lon + 360
    elif (
        city_1_lon > 0 > city_2_lon
        and np.abs(city_1_lon) > 90
        and np.abs(city_2_lon) > 90
    ):
        city_2_lon = city_2_lon + 360

    fig.add_trace(
        go.Scattergeo(
            locationmode="USA-states",
            mode="lines",
            lon=[city_1_lon, city_2_lon],
            lat=[city_1_lat, city_2_lat],
            marker={"size": 10},
            line=dict(width=1, color="red"),
        )
    )


def plot_cities(
    cities_dict,
    ga_obj,
    lon=lambda x: x["lon"],
    lat=lambda x: x["lat"],
    name=lambda x: x["Location"] + ", " + x["State"],
    scope=None,
    projection_type="mercator",
):

    lons = [lon(city) for city in cities_dict.values()]
    lats = [lat(city) for city in cities_dict.values()]
    names = [name(city) for city in cities_dict.values()]

    fig = go.Figure()

    fig.add_trace(
        go.Scattergeo(
            locationmode="USA-states",
            lon=lons,
            lat=lats,
            hoverinfo="text",
            text=names,
            mode="markers",
            marker=dict(
                size=2,
                color="rgb(255, 0, 0)",
                line=dict(width=3, color="rgba(68, 68, 68, 0)"),
            ),
        )
    )

    geo = dict(
        projection_type=projection_type,
        showland=True,
        landcolor="rgb(243, 243, 243)",
        countrycolor="rgb(204, 204, 204)",
    )

    if scope:
        geo.update({"scope": scope})

    fig.update_layout(
        title_text="Best Route", showlegend=False, geo=geo,
    )

    for city_1, city_2 in zip(ga_obj.best_individual_, ga_obj.best_individual_[1:]):
        add_trace(fig, cities_dict, city_1, city_2, lon, lat)

    add_trace(
        fig,
        cities_dict,
        ga_obj.best_individual_[0],
        ga_obj.best_individual_[-1],
        lon,
        lat,
    )

    fig.show()



import os

import pandas as pd


try:
    module_path = os.path.abspath(os.path.join("."))
    world_capitals = pd.read_csv(os.path.join(module_path, "data/world_capitals.csv"))
except FileNotFoundError:
    module_path = os.path.abspath(os.path.join(".."))
    world_capitals = pd.read_csv(os.path.join(module_path, "data/world_capitals.csv"))


world_capitals.dropna(subset=["CapitalName"], axis=0, inplace=True)

world_capitals.reset_index(drop=True, inplace=True)
world_capitals.index += 1

world_capitals_dict = world_capitals.to_dict(orient="index")



exception_messages = {
    "InvalidSelectionStrategy": lambda selection_strategy, allowed_selection_strategies: f"{selection_strategy} is not a valid selection strategy. "
    f"Available options are {', '.join(allowed_selection_strategies)}.",
    "InvalidPopulationSize": "The population size must be larger than 2",
    "InvalidExcludedGenes": lambda excluded_genes: f"{excluded_genes} is not a valid input for excluded_genes",
}



import random
import time
from functools import reduce
import hashlib
from collections import defaultdict

import numpy as np
from numba import njit
from numba.core import types
from numba.typed import Dict

from geneal.applications.tsp.mutation_strategies import MutationStrategies
from geneal.genetic_algorithms import ContinuousGenAlgSolver
from geneal.utils.exceptions import InvalidInput


mutation_options = {"random_swap", "random_inversion", "2-opt"}

allowed_mutations = {
    "2-opt",
    "random_swap",
    "random_inversion",
    "random_gene_nearest_neighbour",
    "worst_gene_random",
    "worst_gene_nearest_neighbour",
    "select_any_mutation",
}


@njit
def fitness_function(individual, edges):
    """
    Implements the logic that calculates the fitness
    measure of an individual. It sums all the costs of going
    from node to node in the tour.

    :param individual: chromosome of genes representing an individual
    :param edges: dictionary with cost between all nodes
    :return: the fitness of the individual
    """

    total_length = 0
    for i in range(individual.shape[0] - 1):
        total_length += edges[(individual[i], individual[i + 1])]

    total_length += edges[(individual[0], individual[-1])]

    return -round(total_length, 2)


def convert_to_typed_dict(G):

    edges_dict = Dict.empty(
        key_type=types.UniTuple(types.int64, 2), value_type=types.float64
    )

    edges_dict.update({(edge[1], edge[0]): G.edges[edge]["weight"] for edge in G.edges})
    edges_dict.update({(edge[0], edge[1]): G.edges[edge]["weight"] for edge in G.edges})

    return edges_dict


class TravellingSalesmanProblemSolver(MutationStrategies, ContinuousGenAlgSolver):
    def __init__(
        self,
        graph,
        mutation_strategy: str = "2-opt",
        n_searches: int = 1,
        numba_speedup: bool = False,
        *args,
        **kwargs,
    ):

        self.check_input(kwargs, graph)

        MutationStrategies.__init__(self, n_searches=n_searches)
        ContinuousGenAlgSolver.__init__(self, n_crossover_points=2, *args, **kwargs)

        if mutation_strategy not in allowed_mutations:
            raise (InvalidInput(f"{mutation_strategy} is an invalid mutation strategy"))

        if numba_speedup:

            edges_dict = convert_to_typed_dict(graph)

            self.fitness_function = lambda individual: fitness_function(
                individual, edges_dict
            )

        self.G = graph
        self.mutation_strategy = mutation_strategy
        self.fitness_time = 0
        self.chromosomes = defaultdict(int)

    @staticmethod
    def check_input(kwargs, graph):
        if "n_crossover_points" in kwargs:

            if kwargs["n_crossover_points"] != 2:
                print("Defaulting 'n_crossover_points' to 2")
            kwargs.pop("n_crossover_points")

        if "n_genes" in kwargs:
            if kwargs["n_genes"] > len(graph.nodes):
                print(
                    f"'n_genes' can't be larger than the nodes in the graph. The number of genes "
                    f"will default to {len(graph.nodes)}."
                )

                kwargs["n_genes"] = len(graph.nodes)

        else:
            kwargs["n_genes"] = len(graph.nodes)

        return kwargs

    def fitness_function(self, individual):
        """
        Implements the logic that calculates the fitness
        measure of an individual. It sums all the costs of going
        from node to node in the tour.

        :param individual: chromosome of genes representing an individual
        :return: the fitness of the individual
        """

        start_time = time.time()

        arr_hash = hashlib.sha1(individual).hexdigest()

        if arr_hash in self.chromosomes:
            res = self.chromosomes[arr_hash]

        else:
            res = reduce(
                lambda total_length, city_pair: total_length
                + self.G.edges[(city_pair[0], city_pair[1])]["weight"],
                zip(individual, individual[1:]),
                0,
            )

            res += self.G.edges[(individual[0], individual[-1])]["weight"]

            res = -round(res, 2)

            self.chromosomes[arr_hash] = res

        self.fitness_time += time.time() - start_time

        return res

    def initialize_population(self):
        """
        Initializes the population of the problem. It creates a
        matrix of size (pop_size x n_genes) containing permutations of the nodes
        on each row.

        :return: a numpy array with a randomized initialized population
        """

        population = np.repeat(
            np.arange(1, self.n_genes + 1)[np.newaxis, :], self.pop_size, axis=0
        )

        return np.array(list(map(lambda x: np.random.permutation(x), population)))

    def create_offspring(self, first_parent, sec_parent, crossover_pt, _):
        """
        Creates an offspring from 2 parents. It performs an OX crossover, which
        combines genes from each parent, but maintaining the nodes order of the parents.

        http://www.inf.tu-dresden.de/content/institutes/ki/cl/study/summer14/pssai/slides/GA_for_TSP.pdf

        :param first_parent: first parent's chromosome
        :param sec_parent: second parent's chromosome
        :param crossover_pt: points at which to perform the crossover
        :return: the resulting offspring.
        """

        reordered_sec_parent = np.roll(sec_parent, -crossover_pt[1])

        new_arr = first_parent[crossover_pt[0] : crossover_pt[1]]

        new_arr = np.append(new_arr, reordered_sec_parent)

        _, idx = np.unique(new_arr, return_index=True)

        res = np.roll(new_arr[np.sort(idx)], crossover_pt[0])

        if res.shape[0] != 30:
            a = 1

        return res

    def mutate_population(self, population, n_mutations, **kwargs):
        """
        Mutates the population using a 2-opt rule hybrid. It selects the number of rows
        on which mutation will be applied, and then a applies a local search 2-opt rule
        to those rows.

        :param population: the population at a given iteration
        :param n_mutations: number of mutations to be performed. This number is
        calculated according to mutation_rate, but can be adjusted as needed inside this function
        :return: the mutated population
        """

        adjusted_n_mutations = np.ceil(n_mutations / self.n_genes).astype(int)

        if adjusted_n_mutations == 0:
            return population

        mutation_rows = self.get_mutation_rows(adjusted_n_mutations, population)

        mutation_strategy = self.mutation_strategy
        if "mutation_strategy" in kwargs:
            mutation_strategy = kwargs["mutation_strategy"]

        if mutation_strategy == "2-opt":

            return self.two_opt_mutation(population, mutation_rows)

        elif mutation_strategy == "random_swap":

            mutation_cols = self.get_mutation_cols(adjusted_n_mutations, population)

            return self.random_swap_mutation(population, mutation_rows, mutation_cols)

        elif mutation_strategy == "random_gene_around_nearest_neighbour":

            return self.random_gene_around_nearest_neighbour_mutation(
                population, mutation_rows
            )

        elif mutation_strategy == "random_gene_nearest_neighbour":

            return self.random_gene_nearest_neighbour_mutation(
                population, mutation_rows
            )

        elif mutation_strategy == "worst_gene_random":

            return self.worst_gene_random_mutation(population, mutation_rows)

        elif mutation_strategy == "worst_gene_nearest_neighbour":

            return self.worst_gene_nearest_neighbour_mutation(population, mutation_rows)

        elif mutation_strategy == "random_inversion":

            return self.random_inversion_mutation(
                population,
                mutation_rows,
                np.random.choice(int(population.shape[1] / 2), 1)[0],
            )

        elif mutation_strategy == "select_any_mutation":

            selected_strategy = random.sample(mutation_options, 1)[0]

            return self.mutate_population(
                population, n_mutations, **{"mutation_strategy": selected_strategy}
            )

        else:
            raise (InvalidInput(f"{mutation_strategy} is an invalid mutation strategy"))

    def find_worst_gene(self, chromosome):

        distances = [
            self.G.edges[(chromosome[-1], chromosome[0])]["weight"]
            + self.G.edges[(chromosome[0], chromosome[1])]["weight"],
            *[
                self.G.edges[(city_pair[0], city_pair[1])]["weight"]
                + self.G.edges[(city_pair[1], city_pair[2])]["weight"]
                for city_pair in zip(chromosome, chromosome[1:], chromosome[2:])
            ],
            self.G.edges[(chromosome[-2], chromosome[-1])]["weight"]
            + self.G.edges[(chromosome[-1], chromosome[0])]["weight"],
        ]

        worst_gene = np.argmax(distances)

        return worst_gene



import networkx as nx


def create_graph(
    cities_dict, distance_f, lon=lambda x: x["lon"], lat=lambda x: x["lat"]
):

    G = nx.Graph()

    G.add_nodes_from(cities_dict.keys())
    nx.set_node_attributes(G, cities_dict, "coords")

    for city, city_attr in cities_dict.items():
        edges = [
            (
                city,
                to_city,
                round(
                    distance_f(
                        [lon(city_attr), lat(city_attr)],
                        [lon(to_city_attr), lat(to_city_attr)],
                    ),
                    2,
                ),
            )
            for to_city, to_city_attr in cities_dict.items()
            if to_city != city
        ]

        edges = sorted(edges, key=lambda x: x[2])

        G.add_weighted_edges_from(edges)

    return G



import numpy as np
import networkx as nx
from numba import njit


class MutationStrategies:
    def __init__(self, n_searches: int = 5):
        self.n_searches = n_searches

    def fitness_function(self, individual):
        pass

    def local_search(self, route, method, **kwargs):

        """
        It applies a 2-opt local search to a particular route. It selects 10 combinations
        of nodes on a given tour, performs a 2-opt node flipping,
        calculates the fitness of the resulting new tours and then selects the one
        with maximum fitness

        :param route: a given route to be mutated.
        :param method: method to use to perform local search.
        :return: the mutated route
        """

        mutation_size = kwargs["n_elements"] if "n_elements" in kwargs else 2

        mutation_cols = self.get_consecutive_mutation_cols(
            route, self.n_searches, mutation_size
        )

        local_search_rows = np.array(
            list(
                map(
                    lambda args: method(*args),
                    zip(
                        np.repeat(route[np.newaxis, :], mutation_cols.shape[0], axis=0),
                        mutation_cols,
                    ),
                )
            )
        )

        fitness = np.array(
            list(map(lambda row: self.fitness_function(row), local_search_rows))
        )

        route_fitness = self.fitness_function(route)

        max_fitness = np.argmax(fitness)

        return (
            local_search_rows[max_fitness, :]
            if fitness[max_fitness] > route_fitness
            else route
        )

    def two_opt_mutation(self, population, mutation_rows):

        population[mutation_rows, :] = np.array(
            list(
                map(
                    lambda route: self.two_opt_mutation_helper(route),
                    population[mutation_rows, :],
                )
            )
        )

        return population

    def two_opt_mutation_helper(self, route):

        mutation_cols = self.get_consecutive_mutation_cols(route, 2, 2, replace=False)

        rolled_arr = np.roll(route, route.shape[0] - mutation_cols[1, 1])

        next_index = np.argwhere(rolled_arr == route[mutation_cols[0, 0]])[0, 0]

        return np.hstack(
            (rolled_arr[: next_index + 1], np.flip(rolled_arr[next_index + 1 :]))
        )

    def random_inversion_mutation(self, population, mutation_rows, n_elements):
        """
        Performs a series of random flips and returns route with
        best fitness

        :param population: the current population
        :param mutation_rows: the indexes of the rows to be mutated
        :param n_elements: number of elements to be flipped
        :return: the mutated population
        """

        population[mutation_rows, :] = np.array(
            list(
                map(
                    lambda route: self.local_search(
                        route,
                        self.random_inversion_mutation_helper,
                        **{"n_elements": n_elements}
                    ),
                    population[mutation_rows, :],
                )
            )
        )

        return population

    @staticmethod
    def random_inversion_mutation_helper(route, mutation_cols):
        """
        Flips the nodes between 2 node positions for a given route.

        :param route: a given route to be mutated
        :param mutation_cols: the start position of the nodes to be flipped in between
        :return: the mutated route
        """

        np.put(route, mutation_cols, np.flip(route[mutation_cols]), mode="wrap")

        return route

    def random_gene_nearest_neighbour_mutation(self, population, mutation_rows):
        """
        Random gene inserted beside nearest neighbour mutation (RGIBNNM)

        4.2.7 - https://arxiv.org/pdf/1801.02827.pdf

        :param population: the current population
        :param mutation_rows: the indexes of the rows to be mutated
        :return: the mutated population
        """

        population[mutation_rows, :] = np.array(
            list(
                map(
                    lambda route: self.random_gene_nearest_neighbour_helper(route,),
                    population[mutation_rows, :],
                )
            )
        )

        return population

    def random_gene_nearest_neighbour_helper(self, route):
        """
        Random gene inserted beside nearest neighbour mutation (RGIBNNM)
        helper function

        :param route: current route to be mutated
        :return: the mutated route
        """

        random_gene = np.random.choice(route, 1)[0]

        random_gene_index = np.argwhere(route == random_gene)[0, 0]

        closest_neighbour = list(nx.neighbors(self.G, random_gene))[0]

        closest_neighbour_index = np.argwhere(route == closest_neighbour)[0, 0]

        if random_gene_index < closest_neighbour_index:
            return np.hstack(
                (
                    route[:random_gene_index],
                    route[random_gene_index + 1 : closest_neighbour_index + 1],
                    route[random_gene_index],
                    route[closest_neighbour_index + 1 :],
                )
            )

        else:
            return np.hstack(
                (
                    route[:closest_neighbour_index],
                    route[random_gene_index],
                    route[closest_neighbour_index:random_gene_index],
                    route[random_gene_index + 1 :],
                )
            )

    @staticmethod
    def random_swap_mutation(population, mutation_rows, mutation_cols):
        """
        Performs a swap between elements along rows

        :param population: the current population
        :param mutation_rows: the rows to be mutated
        :param mutation_cols: an matrix containing the indexes of the
        items to be swapped along the rows
        :return: the mutated population
        """

        (
            population[mutation_rows, mutation_cols[:, 0]],
            population[mutation_rows, mutation_cols[:, 1]],
        ) = (
            population[mutation_rows, mutation_cols[:, 1]],
            population[mutation_rows, mutation_cols[:, 0]],
        )

        return population

    @staticmethod
    def swap_genes(route, mutation_cols):
        """
        Performs a swap between elements along rows

        :param population: the current population
        :param mutation_rows: the rows to be mutated
        :param mutation_cols: an matrix containing the indexes of the
        items to be swapped along the rows
        :return: the mutated population
        """

        (route[mutation_cols[0]], route[mutation_cols[1]],) = (
            route[mutation_cols[1]],
            route[mutation_cols[0]],
        )

        return route

    def worst_gene_random_mutation(self, population, mutation_rows):

        """
        Worst left right gene with random gene mutation (WLRGWRGM)

        4.2.3 - https://arxiv.org/pdf/1801.02827.pdf

        :param population: the current population
        :param mutation_rows: the indexes of the rows to be mutated
        :return: the mutated population
        """

        population[mutation_rows, :] = np.array(
            list(
                map(
                    lambda route: self.worst_gene_random_mutation_helper(route,),
                    population[mutation_rows, :],
                )
            )
        )

        return population

    def worst_gene_random_mutation_helper(self, route):
        """
        Worst left right gene with random gene mutation (WLRGWRGM) helper function

        :param route: current route to be mutated
        :return: the mutated route
        """

        worst_gene = self.find_worst_gene(route)

        random_gene = worst_gene
        while random_gene == worst_gene:
            random_gene = np.random.choice(np.arange(route.shape[0]), 1)[0]

        return self.swap_genes(route, np.array([random_gene, worst_gene]))

    def worst_gene_nearest_neighbour_mutation(self, population, mutation_rows):
        """
        Worst gene with nearest neighbour mutation (WGWNNM)

        4.2.4 - https://arxiv.org/pdf/1801.02827.pdf

        :param population: the current population
        :param mutation_rows: the indexes of the rows to be mutated
        :return: the mutated population
        """

        population[mutation_rows, :] = np.array(
            list(
                map(
                    lambda route: self.worst_gene_random_mutation_helper(route,),
                    population[mutation_rows, :],
                )
            )
        )

        return population

    def worst_gene_nearest_neighbour_mutation_helper(self, route):
        """
        Worst gene with nearest neighbour mutation (WGWNNM) helper function

        :param route: current route to be mutated
        :return: the mutated route
        """

        worst_gene = self.find_worst_gene(route)

        closest_neighbour = list(nx.neighbors(self.G, route[worst_gene]))[0]

        closest_neighbour_index = np.argwhere(route == closest_neighbour)[0, 0]

        gene_candidates = np.take(
            np.arange(route.shape[0]),
            np.arange(closest_neighbour_index - 5, closest_neighbour_index + 5 + 1),
            mode="wrap",
        )

        selected_gene = np.random.choice(gene_candidates, 1)[0]

        return self.swap_genes(route, np.array([selected_gene, worst_gene]))

    @staticmethod
    def get_mutation_rows(n_mutations, population):
        """
        Retrieves the mutation rows for the current iteration

        :param n_mutations: number of mutations to be performed
        :param population: the current population
        :return: the mutation rows
        """

        mutation_rows = np.random.choice(
            np.arange(1, population.shape[0]), n_mutations, replace=False
        )

        return mutation_rows

    @staticmethod
    def get_mutation_cols(n_mutations, population):
        """
        Retrieves the mutation columns for the current iteration

        :param n_mutations: number of mutations to be performed
        :param population: the current population
        :return: the mutation columns
        """

        mutation_cols = np.array(
            list(
                map(
                    lambda x: np.sort(
                        np.random.choice(
                            np.arange(population.shape[1]), 2, replace=False
                        )
                    ),
                    range(n_mutations),
                )
            )
        ).astype(int)

        return mutation_cols

    @staticmethod
    def get_consecutive_mutation_cols(route, n_rows, slice_size, replace=True):

        return np.apply_along_axis(
            lambda gene_index: np.take(
                np.arange(route.shape[0]),
                np.arange(gene_index, gene_index + slice_size),
                mode="wrap",
            ),
            1,
            np.random.choice(route.shape[0], size=(n_rows, 1), replace=replace),
        )



import turf

from geneal.applications.tsp.examples.us_cities._us_cities import us_cities_dict
from geneal.applications.tsp.helpers import create_graph

G = create_graph(us_cities_dict, turf.distance)



import turf

from geneal.applications.tsp.examples.world_capitals import world_capitals_dict
from geneal.applications.tsp.helpers import create_graph

G = create_graph(
    world_capitals_dict,
    turf.distance,
    lon=lambda x: x["CapitalLongitude"],
    lat=lambda x: x["CapitalLatitude"],
)
